/*  Copyright 2001-2004 David Abrahams.
 *  Distributed under the Boost Software License, Version 1.0.
 *  (See accompanying file LICENSE_1_0.txt or http://www.boost.org/LICENSE_1_0.txt)
 */

#ifndef SUBST_JG20120722_H
#define SUBST_JG20120722_H

#include "object.h"
#include "regexp.h"

regexp * regex_compile( OBJECT * pattern );

#endif
